var EventSystem = /*#__PURE__*/function () {
  function EventSystem() {
    this.listeners = Object.create(null);
  }

  var _proto = EventSystem.prototype;

  _proto.subscribe = function subscribe(eventname, listener) {
    if (!(eventname in this.listeners)) this.listeners[eventname] = [];
    this.listeners[eventname].push(listener);
  };

  _proto.unsubscribe = function unsubscribe(eventname, listener) {
    if (!(eventname in this.listeners)) return;
    this.listeners[eventname] = this.listeners[eventname].filter(function (l) {
      return l !== listener;
    });
  };

  _proto.dispatch = function dispatch(eventname, event) {
    if (!(eventname in this.listeners)) return;
    this.listeners[eventname].forEach(function (listener) {
      return listener(event);
    });
  };

  _proto.clone = function clone() {
    var evSys = new EventSystem();

    for (var _i = 0, _Object$keys = Object.keys(this.listeners); _i < _Object$keys.length; _i++) {
      var _this$listeners$event;

      var event = _Object$keys[_i];
      evSys.listeners[event] = (_this$listeners$event = this.listeners[event]) == null ? void 0 : _this$listeners$event.slice();
    }

    return evSys;
  };

  return EventSystem;
}();

var bind = function bind(fn, self) {
  return fn.bind(self);
};

var rate2duration = function rate2duration(r) {
  return 1000 / r;
};

var duration2rate = function duration2rate(d) {
  return 1000 / d;
};

function multiTimer() {
  var ev = new EventSystem();
  var _steps = [];
  var _stepIndex = 0;

  var _startTime = Date.now();

  var _currStepElapsed = 0;
  var _prevStepElapsed = 0;
  var _shouldCancelFrame = false;
  var _state = 'paused';
  var _tickRate = 60;
  var timer = {};

  var activeStep = function activeStep() {
    return _steps[_stepIndex];
  };

  var updatePrevElapsed = function updatePrevElapsed() {
    return _prevStepElapsed = _steps.slice(0, _stepIndex).reduce(function (acc, s) {
      return acc + s.duration;
    }, 0);
  };

  var event = function event() {
    return {
      timer: timer,
      elapsed: _currStepElapsed,
      totalElapsed: _currStepElapsed + _prevStepElapsed,
      step: activeStep(),
      state: _state,
      stepIndex: _stepIndex
    };
  };

  function steps(steps) {
    if (typeof steps === 'undefined') return _steps;
    _steps = steps;
    return timer;
  }

  function tickRate(rate) {
    if (typeof rate === 'undefined') return _tickRate;
    _tickRate = rate;
    return timer;
  }

  function tickDuration(duration) {
    if (typeof duration === 'undefined') return rate2duration(_tickRate);
    _tickRate = duration2rate(duration);
    return timer;
  }

  var pushStep = function pushStep(step) {
    _steps.push(step);

    return timer;
  };

  function handleTick() {
    if (_shouldCancelFrame) {
      _shouldCancelFrame = false;
      return;
    }

    _currStepElapsed = Date.now() - _startTime;
    var step = activeStep();
    var missingTime = step.duration - _currStepElapsed;
    var isLastStep = _stepIndex >= _steps.length - 1;
    var isStepEnded = missingTime <= 0;
    var isTimerEnded = isStepEnded && isLastStep;

    if (isTimerEnded) {
      // Timer has finished last step
      handleEnd();
      ev.dispatch('end', event());
    } else if (isStepEnded) {
      // Timer has finished current step. Set next stap as active
      // and continue
      handleStepChange();
      ev.dispatch('stepchange', event());
    }

    ev.dispatch('tick', event());

    if (!isTimerEnded) {
      var nextTickDuration = Math.min(rate2duration(_tickRate), missingTime);
      setTimeout(handleTick, nextTickDuration);
    }
  } // Update state when step changes


  function handleStepChange() {
    _stepIndex++;
    _prevStepElapsed = updatePrevElapsed();
    _currStepElapsed -= _prevStepElapsed;
    _startTime = Date.now() - _currStepElapsed;
  } // Update state when timer ends all steps


  function handleEnd() {
    _state = 'stopped';
    _prevStepElapsed = updatePrevElapsed();
    _currStepElapsed = _steps[_stepIndex].duration;
  }

  var start = function start() {
    if (_state === 'playing') {
      process.env.NODE_ENV !== "production" && console.warn('Trying to start a MultiTimer that is already playing, skipping.');
      return;
    }

    _state = 'playing';
    ev.dispatch('start', event());
    _startTime = Date.now() - _currStepElapsed;
    handleTick();
  };

  var pause = function pause() {
    if (_state === 'paused') {
      process.env.NODE_ENV !== "production" && console.warn('Trying to pause a MultiTimer that is already paused, skipping.');
      return;
    }

    _state = 'paused';
    ev.dispatch('pause', event());
    _shouldCancelFrame = true;
  };

  Object.assign(timer, {
    activeStep: activeStep,
    pushStep: pushStep,
    steps: steps,
    start: start,
    pause: pause,
    tickRate: tickRate,
    tickDuration: tickDuration,
    state: function state() {
      return _state;
    },
    on: bind(ev.subscribe, ev),
    off: bind(ev.unsubscribe, ev)
  });
  return timer;
}

export { multiTimer };
//# sourceMappingURL=multi-timer.esm.js.map
